"""
App-independent UNIT tests for underlying approval process support classes - no app logic
"""
import django_fsm as fsm
from django.db import models
from django.test import TestCase

import signoffs.core.signing_order as so
from signoffs.core import process as signoffs_process
from signoffs.core.models.fields import ApprovalField, ApprovalSignoffSingle
from signoffs.core.tests import fixtures
from signoffs.core.tests.models import AbstractLeaveApproval, ApprovalSignoff
from signoffs.registry import register

# Basic tests - no models, no state changes


@register(id="test.process.application")
class Application(AbstractLeaveApproval):
    employee_signoff = ApprovalSignoffSingle(
        ApprovalSignoff.register(id="test.process.application.employee_signoff")
    )


@register(id="test.process.authorization")
class Authorization(AbstractLeaveApproval):
    hr_signoff_type = ApprovalSignoff.register(
        id="test.process.authorization.hr_signoff"
    )
    mngmt_signoff_type = ApprovalSignoff.register(
        id="test.process.authorization.mngmt_signoff"
    )


class ApprovalTransitionTests(TestCase):
    def test_approval_transition(self):
        t = signoffs_process.ApprovalTransition(
            approval_id="test.process.application",
            approve=self.test_approval_transition,
        )
        self.assertEqual(t.approval_id, "test.process.application")
        self.assertEqual(t.approve_name, "test_approval_transition")
        self.assertIsNone(t.revoke)
        self.assertEqual(t.revoke_name, "")

    def test_approval_transition_registry(self):
        r = signoffs_process.ApprovalTransitionRegistry()
        r.add_approval(
            Application
        )  # adding entry sets the ordering, even if no transitions given
        r.add_approve_transition(
            Authorization, self.test_approval_transition_registry
        )  # add transition to new item
        r.add_approve_transition(
            Application, self.test_approval_transition
        )  # go back an add transition to existing
        r.add_revoke_transition(Authorization, self.test_approval_transition_registry)
        self.assertEqual(
            r.get(Authorization).approve, self.test_approval_transition_registry
        )
        self.assertEqual(
            r.get(Authorization).revoke, self.test_approval_transition_registry
        )
        self.assertEqual(len(r.approval_order()), 2)
        self.assertListEqual(r.approval_order(), [Application.id, Authorization.id])
        self.assertListEqual(
            [t.approve for t in r],
            [self.test_approval_transition, self.test_approval_transition_registry],
        )  # Test iter


class ProtoProcessModel:
    """The inkling of the components an approval process model will be made up of"""

    # introspection logic in BoundApprovalSequence requires class variables with ApprovalType same name as instance variables
    # Normally this is handled by declaring these related approvals using an ApprovalField or RelatedApprovalDescriptor
    #  --> this is not a useful code pattern - for unit testing only!
    approval1 = Application
    approval2 = Authorization

    def __init__(self):
        # normally approval instance would be generated by a RelatedApprovalDescriptor
        self.approval1 = Application.create()
        self.approval2 = Authorization.create()

        # actual processes would not manage the transition registry themselves - this is for unit testing only
        self.reg = signoffs_process.ApprovalTransitionRegistry()
        self.reg.add_approve_transition(self.approval1, self.transition1)
        self.reg.add_approve_transition(self.approval2, self.transition2)
        self.reg.add_revoke_transition(self.approval1, self.revoke_transition)

    def transition1(self, approval):
        print("Transition 1 for ", approval)

    def transition2(self, approval):
        print("Transition 2 for ", approval)

    def revoke_transition(self, approval):
        print("Revoke Transition for ", approval)

    def get_approval_sequence(self):
        return signoffs_process.BoundApprovalSequence(
            self, ordering=self.reg.approval_order()
        )

    def get_approval_process(self):
        return signoffs_process.BasicApprovalProcess(
            self, self.reg, approval_sequence=self.get_approval_sequence()
        )


class ApprovalSequenceTests(TestCase):
    def test_transition_registry(self):
        model = ProtoProcessModel()
        self.assertEqual(model.reg.get(model.approval1).approve, model.transition1)
        self.assertIsNone(model.reg.get("NotExists"))

    def test_approval_sequence(self):
        model = ProtoProcessModel()
        seq = model.get_approval_sequence()
        self.assertDictEqual(
            seq, {"approval1": model.approval1, "approval2": model.approval2}
        )


class BasicApprovalTransitionsRegistryTests(TestCase):
    def setUp(self):
        super().setUp()
        model = ProtoProcessModel()
        self.process = model.get_approval_process()
        self.model = model

    def test_approval_sequence(self):
        self.assertDictEqual(
            self.process.seq,
            {"approval1": self.model.approval1, "approval2": self.model.approval2},
        )

    def test_get_all_approvals(self):
        self.assertListEqual(
            self.process.get_all_approvals(),
            [self.model.approval1, self.model.approval2],
        )

    def test_get_approved_approvals(self):
        self.assertListEqual(self.process.get_approved_approvals(), [])

    def test_get_unapproved_approvals(self):
        self.assertListEqual(
            self.process.get_unapproved_approvals(),
            [self.model.approval1, self.model.approval2],
        )

    def test_get_next_approval(self):
        self.assertEqual(self.process.get_next_approval(), self.model.approval1)

    def test_next_approval_is_signed(self):
        self.assertFalse(self.process.next_approval_is_signed())

    def test_get_previous_approval(self):
        self.assertEqual(self.process.get_previous_approval(), None)

    def test_get_available_approvals(self):
        self.assertListEqual(
            self.process.get_available_approvals(),
            [
                self.model.approval1,
            ],
        )

    def test_get_next_available_approval(self):
        self.assertEqual(
            self.process.get_next_available_approval(), self.model.approval1
        )

    def test_get_revokable_approvals(self):
        self.assertListEqual(self.process.get_revokable_approvals(), [])

    def test_get_next_revokable_approval(self):
        self.assertEqual(self.process.get_next_revokable_approval(), None)

    def test_bound_approve_transition(self):
        self.assertEqual(
            self.process.bound_approve_transition(self.model.approval1),
            self.model.transition1,
        )
        self.assertEqual(
            self.process.bound_approve_transition(self.model.approval2),
            self.model.transition2,
        )

    def test_bound_revoke_transition(self):
        self.assertEqual(
            self.process.bound_revoke_transition(self.model.approval1),
            self.model.revoke_transition,
        )
        self.assertEqual(
            self.process.bound_revoke_transition(self.model.approval2), None
        )


# Tests requiring a model with state changes


@register(id="test.process.approval1")
class Approval1(AbstractLeaveApproval):
    signoff1 = ApprovalSignoffSingle(
        ApprovalSignoff.register(id="test.process.approval1.signoff1")
    )

    def is_complete(self):
        """Customize logic for determining if Approval signoffs are complete and ready to be approved."""
        return self.signoff1.get().is_signed()

    def sign_approval(self, user):
        if self.signoff1.can_sign(user):
            self.signoff1.create(user=user)


@register(id="test.process.approval2")
class Approval2(AbstractLeaveApproval):
    signoff1 = ApprovalSignoff.register(id="test.process.approval2.signoff1")
    signoff2 = ApprovalSignoff.register(id="test.process.approval2.signoff2")

    signing_order = so.SigningOrder(so.AtLeastN(signoff1, n=1), signoff2)

    def sign_approval(self, user):
        next = self.next_signoffs(for_user=user)
        if next and not self.is_approved() and not self.has_signed(user):
            next[-1].sign(
                user
            )  # Arbitrarily approve "most nextest" signoff on the approval


class NonFsmApprovalProcessModel(models.Model):
    """An Approval Process model where transitions and approval sequencing are handled by app logic - no ordering"""

    class States(models.TextChoices):
        STATE0 = "Initiated"
        STATE1 = "State 1"
        STATE2 = "State 2"

    state = str(States.STATE0)

    approval1, approval1_stamp = ApprovalField(Approval1)
    approval2, approval2_stamp = ApprovalField(Approval2)

    process = (
        signoffs_process.ApprovalsProcess()
    )  # optionally list approvals in order, default is order they are registered below

    # Decorate approval transitions the long (but flexible) way

    @process.register_approve_transition(approval1)
    @process.do_approval
    def transition1(self, approval):
        self.state = self.States.STATE1
        print("Transition 1", self.state, self.approval1)

    @process.register_revoke_transition(approval1)
    @process.do_revoke
    def revoke1(self, approval):
        self.state = self.States.STATE0
        print("Revoke Transition 1", self.state, self.approval1)

    # or using convenience decorators

    @process.register_and_do_approval(approval2)
    def transition2(self, approval):
        self.state = self.States.STATE2
        print("Transition 2", self.state, self.approval2)

    @process.register_and_do_revoke(approval2)
    def revoke2(self, approval):
        self.state = self.States.STATE1
        print("Revoke Transition 2", self.state, self.approval2)

    def sign_and_approve_approval1(self):
        u = fixtures.get_user()
        self.approval1.sign_approval(u)
        self.process.try_approve_transition(self.approval1, u)

    def sign_and_approve_approval2(self):
        u1 = fixtures.get_user()
        self.approval2.sign_approval(u1)
        u2 = fixtures.get_user()
        self.approval2.sign_approval(u2)
        self.process.try_approve_transition(self.approval2, u2)


class NonFsmApprovalActionsTests(TestCase):
    def setUp(self):
        super().setUp()
        self.model = NonFsmApprovalProcessModel.objects.create()

    def test_make_transition(self):
        # test that the "hard way works"
        u = fixtures.get_user()
        self.model.approval1.sign_approval(u)
        save = self.model.transition1(self.model.approval1)
        self.assertIsInstance(save, self.model.process.transition_save_class)
        save()
        refetch = NonFsmApprovalProcessModel.objects.get(pk=self.model.pk)
        self.assertTrue(refetch.approval1.is_approved())

    def test_can_proceed(self):
        self.assertTrue(self.model.process.can_proceed(self.model.approval1))
        self.assertFalse(self.model.process.can_proceed(self.model.approval2))

        self.model.sign_and_approve_approval1()
        self.assertFalse(self.model.process.can_proceed(self.model.approval1))
        self.assertTrue(self.model.process.can_proceed(self.model.approval2))

    def test_user_can_proceed(self):
        u = fixtures.get_user()
        self.assertTrue(self.model.process.user_can_proceed(self.model.approval1, u))
        self.assertFalse(self.model.process.user_can_proceed(self.model.approval2, u))

    def test_can_do_approve_transition(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.can_do_approve_transition(self.model.approval1, u)
        )

        self.model.approval1.sign_approval(u)
        self.assertTrue(
            self.model.process.can_do_approve_transition(self.model.approval1, u)
        )
        self.assertFalse(
            self.model.process.can_do_approve_transition(self.model.approval2, u)
        )

    def test_is_revokable(self):
        self.assertFalse(self.model.process.is_revokable(self.model.approval1))
        self.assertFalse(self.model.process.is_revokable(self.model.approval2))

        self.model.sign_and_approve_approval1()
        self.assertTrue(self.model.process.is_revokable(self.model.approval1))

    def test_has_revoke_transition_perm(self):
        u = fixtures.get_user()
        self.assertTrue(
            self.model.process.has_revoke_transition_perm(self.model.approval1, u)
        )
        self.assertTrue(
            self.model.process.has_revoke_transition_perm(self.model.approval2, u)
        )

    def test_user_can_revoke(self):
        u = fixtures.get_user()
        self.assertFalse(self.model.process.user_can_revoke(self.model.approval1, u))
        self.assertFalse(self.model.process.user_can_revoke(self.model.approval2, u))

    def test_can_do_revoke_transition(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.can_do_revoke_transition(self.model.approval1, u)
        )

        self.model.sign_and_approve_approval1()
        self.assertTrue(
            self.model.process.can_do_revoke_transition(self.model.approval1, u)
        )
        self.assertFalse(
            self.model.process.can_do_revoke_transition(self.model.approval2, u)
        )

    def test_try_approve_transition(self):
        u = fixtures.get_user()
        u2 = fixtures.get_user()
        self.assertFalse(
            self.model.process.try_approve_transition(self.model.approval1, u)
        )

        self.model.approval1.sign_approval(u)
        self.assertTrue(
            self.model.process.try_approve_transition(self.model.approval1, u)
        )
        self.assertEqual(self.model.state, self.model.States.STATE1)

        self.model.approval2.sign_approval(u)
        self.assertFalse(
            self.model.process.try_approve_transition(self.model.approval2, u)
        )
        self.model.approval2.sign_approval(u2)
        self.assertTrue(
            self.model.process.try_approve_transition(self.model.approval2, u2)
        )
        self.assertEqual(self.model.state, self.model.States.STATE2)

    def test_try_revoke_transition(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.try_revoke_transition(self.model.approval1, u)
        )

        self.model.sign_and_approve_approval1()
        self.model.sign_and_approve_approval2()
        self.assertEqual(self.model.state, self.model.States.STATE2)

        self.assertFalse(
            self.model.process.try_revoke_transition(self.model.approval1, u)
        )
        self.assertTrue(
            self.model.process.try_revoke_transition(self.model.approval2, u)
        )
        self.assertEqual(self.model.state, self.model.States.STATE1)

        self.assertFalse(
            self.model.process.try_revoke_transition(self.model.approval2, u)
        )
        self.assertTrue(
            self.model.process.try_revoke_transition(self.model.approval1, u)
        )
        self.assertEqual(self.model.state, self.model.States.STATE0)


# FSM Approval Process TODO: add transition conditions and permissions to show how this changes results


class FsmApprovalProcessModel(models.Model):
    """An Approval Process model where transitions and approval sequencing are defined by FSM"""

    class States(models.TextChoices):
        STATE0 = "Initiated"
        STATE1 = "State 1"
        STATE2 = "State 2"

    state = fsm.FSMField(choices=States.choices, default=States.STATE0)

    approval1, approval1_stamp = ApprovalField(Approval1)
    approval2, approval2_stamp = ApprovalField(Approval2)

    process = signoffs_process.FsmApprovalsProcess()

    # Decorate FSM approval transitions the long (but flexible) way

    @process.register_approve_transition(approval1)
    @fsm.transition(state, source=States.STATE0, target=States.STATE1)
    @process.do_approval
    def transition1(self, approval):
        print("FSM Transition 1", self.state, self.approval1)

    @process.register_revoke_transition(approval1)
    @fsm.transition(state, source=States.STATE1, target=States.STATE0)
    @process.do_revoke
    def revoke1(self, approval):
        print("FSM Revoke Transition 1", self.state, self.approval1)

    # Or using convenience decorators

    @process.approval_transition(
        approval2, state, source=States.STATE1, target=States.STATE2
    )
    def transition2(self, approval):
        print("FSM Transition 2", self.state, self.approval2)

    @process.revoke_transition(
        approval2, state, source=States.STATE2, target=States.STATE1
    )
    def revoke2(self, approval):
        print("FSM Revoke Transition 2", self.state, self.approval2)

    def sign_and_approve_approval1(self):
        # the convenient way
        u = fixtures.get_user()
        self.approval1.sign_approval(u)
        self.process.try_approve_transition(self.approval1, u)

    def sign_and_approve_approva2(self):
        u1 = fixtures.get_user()
        self.approval2.sign_approval(u1)
        u2 = fixtures.get_user()
        self.approval2.sign_approval(u2)
        self.process.try_approve_transition(self.approval2, u2)


class FsmApprovalActionsTests(TestCase):
    def setUp(self):
        super().setUp()
        self.model = FsmApprovalProcessModel.objects.create()

    def test_make_transition(self):
        # test that the "hard way works"
        u = fixtures.get_user()
        self.model.approval1.sign_approval(u)
        save = self.model.transition1(self.model.approval1)
        self.assertIsInstance(save, self.model.process.transition_save_class)
        save()
        refetch = FsmApprovalProcessModel.objects.get(pk=self.model.pk)
        self.assertTrue(refetch.approval1.is_approved())

    def test_can_proceed(self):
        self.assertTrue(self.model.process.can_proceed(self.model.approval1))
        self.assertFalse(self.model.process.can_proceed(self.model.approval2))

        self.model.sign_and_approve_approval1()
        self.assertFalse(self.model.process.can_proceed(self.model.approval1))
        self.assertTrue(self.model.process.can_proceed(self.model.approval2))

    def test_user_can_proceed(self):
        u = fixtures.get_user()
        self.assertTrue(self.model.process.user_can_proceed(self.model.approval1, u))
        self.assertFalse(self.model.process.user_can_proceed(self.model.approval2, u))

    def test_can_do_approve_transition(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.can_do_approve_transition(self.model.approval1, u)
        )

        self.model.approval1.sign_approval(u)
        self.assertTrue(
            self.model.process.can_do_approve_transition(self.model.approval1, u)
        )
        self.assertFalse(
            self.model.process.can_do_approve_transition(self.model.approval2, u)
        )

    def test_is_revokable(self):
        self.assertFalse(self.model.process.is_revokable(self.model.approval1))
        self.assertFalse(self.model.process.is_revokable(self.model.approval2))

        self.model.sign_and_approve_approval1()
        self.assertTrue(self.model.process.is_revokable(self.model.approval1))

    def test_has_revoke_transition_perm(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.has_revoke_transition_perm(self.model.approval1, u)
        )
        self.assertFalse(
            self.model.process.has_revoke_transition_perm(self.model.approval2, u)
        )

        self.model.sign_and_approve_approval1()
        self.assertTrue(
            self.model.process.has_revoke_transition_perm(self.model.approval1, u)
        )

    def test_user_can_revoke(self):
        u = fixtures.get_user()
        self.assertFalse(self.model.process.user_can_revoke(self.model.approval1, u))
        self.assertFalse(self.model.process.user_can_revoke(self.model.approval2, u))

    def test_can_do_revoke_transition(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.can_do_revoke_transition(self.model.approval1, u)
        )

        self.model.sign_and_approve_approval1()
        self.assertTrue(
            self.model.process.can_do_revoke_transition(self.model.approval1, u)
        )
        self.assertFalse(
            self.model.process.can_do_revoke_transition(self.model.approval2, u)
        )

    def test_try_approve_transition(self):
        u = fixtures.get_user()
        u2 = fixtures.get_user()
        self.assertFalse(
            self.model.process.try_approve_transition(self.model.approval1, u)
        )

        self.model.approval1.sign_approval(u)
        self.assertTrue(
            self.model.process.try_approve_transition(self.model.approval1, u)
        )
        self.assertEqual(self.model.state, self.model.States.STATE1)

        self.model.approval2.sign_approval(u)
        self.assertFalse(
            self.model.process.try_approve_transition(self.model.approval2, u)
        )
        self.model.approval2.sign_approval(u2)
        self.assertTrue(
            self.model.process.try_approve_transition(self.model.approval2, u2)
        )
        self.assertEqual(self.model.state, self.model.States.STATE2)

    def test_try_revoke_transition(self):
        u = fixtures.get_user()
        self.assertFalse(
            self.model.process.try_revoke_transition(self.model.approval1, u)
        )

        self.model.sign_and_approve_approval1()
        self.model.sign_and_approve_approva2()
        self.assertEqual(self.model.state, self.model.States.STATE2)

        self.assertFalse(
            self.model.process.try_revoke_transition(self.model.approval1, u)
        )
        self.assertTrue(
            self.model.process.try_revoke_transition(self.model.approval2, u)
        )
        self.assertEqual(self.model.state, self.model.States.STATE1)

        self.assertFalse(
            self.model.process.try_revoke_transition(self.model.approval2, u)
        )
        self.assertTrue(
            self.model.process.try_revoke_transition(self.model.approval1, u)
        )
        self.assertEqual(self.model.state, self.model.States.STATE0)



"""
    All Behavioural "Types" are loaded in a global registry to they can be accessed anywhere.
"""

from django.core.exceptions import ImproperlyConfigured
from persisting_theory import Registry


class ObjectRegistry(Registry):
    """Generic base class for efficiently registering a bunch of objects that have an id attribute to use as name"""

    object_type = object
    name_attr = "id"

    def validate(self, data):
        """Return True iff the data can is a unique, vaild candidate for storage in this registry"""
        class_validator = getattr(data, "validate", lambda: True)
        return (
            issubclass(data, self.object_type)
            and not getattr(data, self.name_attr) in self
            and class_validator()
        )

    def prepare_name(self, data, name=None):
        """Name (key) in registry will ordinarily be data.id, but can be overridden"""
        return name or getattr(data, self.name_attr)

    def register(self, *data, name=None, **kwargs):
        """Allow for multiple objects to be registered at once, name is ignored, obj.name_attr used instead"""
        for obj in data:
            super().register(obj, **kwargs)


class SignoffTypes(ObjectRegistry):
    """Keep a reference to all Signoff Types"""

    look_into = "signoffs"

    @property
    def object_type(self):
        """defer dependency to prevent cyclical imports"""
        import signoffs.core.signoffs

        return signoffs.core.signoffs.AbstractSignoff


signoffs = SignoffTypes()
"""Singleton - the Signoff Types registry. `(see persisting_theory.Registry)`"""


def get_signoff_type(signoff_id_or_type):
    """
    Return a registered Signoff Type or raise ImproperlyConfigured if no such type was registered.
    Convenience function accepts either a Type or an id, and checks for existence.
    """
    signoff_type = (
        signoffs.get(signoff_id_or_type)
        if isinstance(signoff_id_or_type, str)
        else signoff_id_or_type
    )
    if signoff_type is None:
        raise ImproperlyConfigured(
            f"Signoff Type {signoff_type} must be registered before it can be used."
        )
    return signoff_type


class ApprovalTypes(ObjectRegistry):
    """Keep a reference to all Approval Types"""

    look_into = "approvals"

    @property
    def object_type(self):
        """defer dependency to prevent cyclical imports"""
        import signoffs.core.approvals

        return signoffs.core.approvals.AbstractApproval


approvals = ApprovalTypes()
"""Singleton - the Approval Types registry. `(see persisting_theory.Registry)`"""


def get_approval_type(approval_id_or_type):
    """
    Return a registered Approval Type or raise ImproperlyConfigured if not such type was registered.
    Convenience function accepts either a Type or an id, and checks for existence.
    """
    approval_type = (
        approvals.get(approval_id_or_type)
        if isinstance(approval_id_or_type, str)
        else approval_id_or_type
    )
    if approval_type is None:
        raise ImproperlyConfigured(
            f"Approval Type {approval_id_or_type} must be registered before it can be used."
        )
    return approval_type


def get_approval_id(approval_id_or_type):
    """
    Return the str approval.id for an approval, approval_type, or approval_id object.
    """
    return (
        approval_id_or_type
        if isinstance(approval_id_or_type, str)
        else approval_id_or_type.id
    )


# Class decorator to simplify registering a base Type class


def register(id, **kwargs):
    """
    Return a class decorator to register a "Type" for the given (e.g., Signoff or Approval) class with the given id

    class to be registered MUST have a "register(id, **kwargs)" method
    The decorator returns the registered class in place of the decorated class

    Usage::

        @register(id='myapp.approval')
        class MyApproval(BaseApproval):
            # ...
    """

    def decorator(cls):
        return cls.register(id=id, **kwargs)

    return decorator


__all__ = ["signoffs", "approvals", "register"]



"""
    Test Suite for regex_match pattern matching language
"""
from types import SimpleNamespace

from django.test import SimpleTestCase

from ..signoff_pattern import (
    AnyOneOf,
    AtLeastN,
    ExactlyN,
    ExactlyOne,
    InParallel,
    InSeries,
    OneOrMore,
    Optional,
    PatternSet,
    SigningOrderPattern,
    ZeroOrMore,
)

#
#  Pluggable encode/decode logic for pattern objects and token objects
#
# For test cases:
#   Pattern objects are classes and tokens are instances of those classes
obj_repr = SimpleNamespace(
    pattern_to_str=lambda obj: obj.__name__,
    pattern_from_str=lambda name: globals()[name]
    if name in globals()
    else __builtins__[name],
    to_str=lambda obj: type(obj).__name__,
)


class SignoffPatternTests(SimpleTestCase):
    def assertMatch(self, m, is_valid, is_complete, next_tokens):
        self.assertEqual(m.is_valid, is_valid)
        self.assertEqual(m.is_complete, is_complete)
        self.assertSetEqual(set(m.next), set(next_tokens))

    def test_exactly_one(self):
        s = ExactlyOne(str, token_repr=obj_repr)
        self.assertEqual(
            s.terms(),
            {
                str,
            },
        )
        match = s.match("ABC")
        self.assertMatch(match, True, True, [])
        match = s.match(123, "ABC")
        self.assertMatch(match, False, False, [])
        match = s.match()
        self.assertMatch(
            match,
            True,
            False,
            [
                str,
            ],
        )

    def test_optional(self):
        s = Optional(str, token_repr=obj_repr)
        self.assertEqual(
            s.terms(),
            {
                str,
            },
        )
        match = s.match()
        self.assertMatch(
            match, True, False, [str]
        )  # No pattern is complete unless it has at least one token!
        match = s.match("ABC")
        self.assertMatch(match, True, True, [])
        match = s.match(123, "ABC")
        self.assertMatch(match, False, False, [])

    def test_zero_or_more(self):
        s = ZeroOrMore(str, token_repr=obj_repr)
        self.assertEqual(
            s.terms(),
            {
                str,
            },
        )
        match = s.match()
        self.assertMatch(
            match, True, False, [str]
        )  # no pattern is compelte without at least one token
        match = s.match("ABC")
        self.assertMatch(match, True, True, [str])
        match = s.match("ABC", "ABC")
        self.assertMatch(match, True, True, [str])

    def test_series_with_optionals(self):
        s = InSeries(
            ExactlyOne(str, token_repr=obj_repr),
            Optional(str, token_repr=obj_repr),
            ZeroOrMore(int, token_repr=obj_repr),
            token_repr=obj_repr,
        )
        self.assertEqual(s.terms(), {str, int})
        match = s.match("ABC")
        self.assertMatch(
            match, True, True, [str, int]
        )  # pattern complete, but optionals are possible
        match = s.match("ABC", "DEF")
        self.assertMatch(match, True, True, [int])
        match = s.match("ABC", "DEF", 123)
        self.assertMatch(match, True, True, [int])
        match = s.match("ABC", "DEF", 123, 456)
        self.assertMatch(match, True, True, [int])
        match = s.match("ABC", 123)
        self.assertMatch(match, True, True, [int])

    def test_one_or_more(self):
        s = OneOrMore(str, token_repr=obj_repr)
        self.assertEqual(
            s.terms(),
            {
                str,
            },
        )
        match = s.match()
        self.assertMatch(match, True, False, [str])
        match = s.match("ABC")
        self.assertMatch(match, True, True, [str])
        match = s.match("ABC", "ABC")
        self.assertMatch(match, True, True, [str])

    def test_at_least_n(self):
        s = AtLeastN(str, n=3, token_repr=obj_repr)
        self.assertEqual(
            s.terms(),
            {
                str,
            },
        )
        match = s.match("ABC", "DEF", "HIJ")
        self.assertMatch(
            match,
            True,
            True,
            [
                str,
            ],
        )
        match = s.match(123, "ABC")
        self.assertMatch(match, False, False, [])
        match = s.match(
            "ABC",
            "DEF",
        )
        self.assertMatch(match, True, False, [str])

    def test_exactly_n(self):
        s = ExactlyN(str, n=3, token_repr=obj_repr)
        self.assertEqual(
            s.terms(),
            {
                str,
            },
        )
        match = s.match("ABC", "DEF", "HIJ")
        self.assertMatch(match, True, True, [])
        match = s.match(123, "ABC")
        self.assertMatch(match, False, False, [])
        match = s.match(
            "ABC",
            "DEF",
        )
        self.assertMatch(match, True, False, [str])
        match = s.match("ABC", "DEF", "HIJ")
        self.assertMatch(match, True, True, [])

    def test_pattern_set(self):
        s = PatternSet(
            AtLeastN(int, n=2, token_repr=obj_repr),
            str,
            AtLeastN(object, n=1, token_repr=obj_repr),
            int,
            token_repr=obj_repr,
        )
        self.assertEqual(len(s), 4)
        self.assertTrue(all(isinstance(t, SigningOrderPattern) for t in s))
        self.assertEqual(
            tuple(type(t) for t in s), (AtLeastN, ExactlyOne, AtLeastN, ExactlyOne)
        )

    def test_in_series(self):
        s = InSeries(
            ExactlyOne(int, token_repr=obj_repr),
            AtLeastN(str, n=2, token_repr=obj_repr),
            AtLeastN(int, n=1, token_repr=obj_repr),
            token_repr=obj_repr,
        )
        self.assertEqual(
            s.terms(),
            {
                int,
                str,
            },
        )
        match = s.match(123, "ABC", "DEF", "HIJ", 456, 789)
        self.assertMatch(match, True, True, [int])
        match = s.match(123, 456, "ABC", "DEF", 789)
        self.assertMatch(match, False, False, [])
        match = s.match(
            123,
            "ABC",
            "DEF",
        )
        self.assertMatch(match, True, False, [str, int])
        match = s.match(123, "ABC", "DEF", "HIJ", 456, 789)
        self.assertMatch(match, True, True, [int])

    def test_in_parallel(self):
        s = InParallel(
            ExactlyOne(int, token_repr=obj_repr),
            AtLeastN(str, n=2, token_repr=obj_repr),
            AtLeastN(object, n=1, token_repr=obj_repr),
            token_repr=obj_repr,
        )
        self.assertEqual(s.terms(), {int, str, object})
        match = s.match(object(), "ABC", "DEF", "HIJ", 456)
        self.assertMatch(match, True, True, [])
        match = s.match(123, "ABC", 456, "DEF", object(), 789)
        self.assertMatch(match, False, False, [])
        match = s.match(
            "ABC",
            "DEF",
            object(),
            object(),
        )
        self.assertMatch(match, True, False, [object, int])

    def test_any_one_of(self):
        s = AnyOneOf(
            ExactlyOne(int, token_repr=obj_repr),
            OneOrMore(str, token_repr=obj_repr),
            ExactlyN(object, n=2, token_repr=obj_repr),
            token_repr=obj_repr,
        )
        self.assertEqual(s.terms(), {int, str, object})
        match = s.match()
        self.assertMatch(match, True, False, [int, str, object])
        match = s.match(
            object(),
        )
        self.assertMatch(match, True, False, [object])
        match = s.match(object(), object())
        self.assertMatch(match, True, True, [])
        match = s.match(42)
        self.assertMatch(match, True, True, [])
        match = s.match("ABC", "DEF")
        self.assertMatch(match, True, True, [str])


# EXAMPLE USAGE


# Some example classes for which objects will be gathered in a Pattern
class A:
    pass


class B:
    pass


class C:
    pass


class SignoffPatternExampleUsageTests(SimpleTestCase):
    # Example Pattern Specification:
    pattern = InSeries(
        InParallel(
            ExactlyOne(A, token_repr=obj_repr),
            AtLeastN(B, n=1, token_repr=obj_repr),
            token_repr=obj_repr,
        ),
        ExactlyOne(C, token_repr=obj_repr),
        AtLeastN(A, n=3, token_repr=obj_repr),
        token_repr=obj_repr,
    )

    def test_pattern_terms(self):
        self.assertEqual(self.pattern.terms(), {A, B, C})

    def assertMatch(self, m, is_valid, is_complete, next_tokens):
        self.assertEqual(m.is_valid, is_valid)
        self.assertEqual(m.is_complete, is_complete)
        self.assertSetEqual(set(m.next), set(next_tokens))

    def test_objective_function(self):
        """given a list of tokens, do the tokens match the pattern?  If not, what is the next token needed?"""

        match = self.pattern.match(B(), B(), A(), C(), A(), A(), A())
        self.assertMatch(
            match,
            True,
            True,
            [
                A,
            ],
        )
        match = self.pattern.match(
            B(),
            B(),
            A(),
            C(),
            A(),
            A(),
        )
        self.assertMatch(
            match,
            True,
            False,
            [
                A,
            ],
        )
        match = self.pattern.match(
            B(),
            A(),
        )
        self.assertMatch(
            match,
            True,
            False,
            [
                C,
            ],
        )
        match = self.pattern.match(
            A(),
            B(),
        )
        self.assertMatch(match, True, False, [B, C])



from .signing_order import (
    SigningOrder,
    SigningOrderPatternMatcher,
    SigningOrderStrategyProtocol,
)
from .signoff_pattern import (
    AnyOneOf,
    AtLeastN,
    ExactlyN,
    ExactlyOne,
    InParallel,
    InSeries,
    OneOrMore,
    Optional,
)



"""
App-independent tests for signoff forms - no app logic
"""
from django.test import SimpleTestCase, TestCase

from signoffs.forms import AbstractSignoffForm, signoff_form_factory

from . import fixtures, models


class FormTestsSignoff(models.BasicSignoff):
    signetModel = models.Signet
    label = "Consent?"


signoff_type = FormTestsSignoff.register(
    id="test.form_signoff", perm="auth.add_signoff"
)


class SignoffFormTests(SimpleTestCase):
    def setUp(self):
        self.formClass = signoff_form_factory(signoff_type=signoff_type)

    def test_form_factory(self):
        self.assertTrue(issubclass(self.formClass, AbstractSignoffForm))

    def test_bound_form(self):
        data = dict(signed_off="True", signoff_id=signoff_type.id)
        bf = self.formClass(data=data)
        self.assertTrue(bf.is_signed_off())

    def test_invalid_bound_form(self):
        data = dict(signoff="True", signoff_id="invalid.type")
        bf = self.formClass(data)
        self.assertFalse(bf.is_valid())


class SignoffWithUserTests(TestCase):
    def setUp(self):
        self.formClass = signoff_form_factory(signoff_type=signoff_type)

    def test_is_signed_off(self):
        data = dict(signed_off="True", signoff_id=signoff_type.id)
        bf = self.formClass(data=data)
        self.assertTrue(bf.is_signed_off())

    def test_sign(self):
        u = fixtures.get_user(perms=("add_signoff",))
        data = dict(signed_off="True", signoff_id=signoff_type.id)
        bf = self.formClass(data=data)
        self.assertTrue(bf.is_valid())
        v = bf.sign(user=u, commit=False)
        self.assertEqual(v.id, signoff_type.id)
        self.assertEqual(v.signet.user, u)

    def test_invalid_signoff(self):
        data = dict(signed_off="True", signoff_id="invalid.type")
        bf = self.formClass(data)
        self.assertFalse(bf.is_valid())



# Generated by Django 3.2.13 on 2022-10-12 19:18

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import signoffs.core.models.signets


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Signet",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "signoff_id",
                    models.CharField(
                        max_length=100,
                        validators=[signoffs.core.models.signets.validate_signoff_id],
                        verbose_name="Signoff Type",
                    ),
                ),
                ("sigil", models.CharField(max_length=256, verbose_name="Signed By")),
                ("sigil_label", models.CharField(max_length=256, null=True)),
                (
                    "timestamp",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="RevokedSignet",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("reason", models.TextField(blank=True, null=True)),
                (
                    "timestamp",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="Revoked at",
                    ),
                ),
                (
                    "signet",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="revoked",
                        to="signoffs_signets.signet",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Revoked by",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



"""
App-dependent tests for signoff template tags
"""
from django.template import Context, Template
from django.test import TestCase

from signoffs.approvals import (
    ApprovalInstanceRenderer,
    ApprovalLogic,
    ApprovalSignoff,
    SimpleApproval,
    signing_order,
)
from signoffs.core.renderers import helpers
from signoffs.core.tests import fixtures
from signoffs.signoffs import SignoffInstanceRenderer, SignoffLogic, utils
from tests.test_app import signoffs


class DummySignoffInstanceRenderer(SignoffInstanceRenderer):
    """Dummy renderer with easy-to-test presentation logic"""

    def signet(self, request_user=None, context=None, **kwargs):
        return (
            f"Signed Signet for {self.signoff.id}" if self.signoff.is_signed() else ""
        )

    def form(self, request_user=None, context=None, **kwargs):
        request_user = helpers.resolve_request_user(request_user, context, **kwargs)
        return (
            f"Signoff Form for {self.signoff.id}"
            if request_user and self.signoff.can_sign(request_user)
            else ""
        )


DummySignoffRenderer = utils.service(DummySignoffInstanceRenderer)


class RenderSignoffTagTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        perm = fixtures.get_perm("can_sign")
        cls.user = fixtures.get_user(perms=(perm,))
        fixtures.get_perm("can_revoke")
        cls.signoff_type = signoffs.TestSignoff.register(
            id="test.signoff",
            logic=SignoffLogic(perm="auth.can_sign", revoke_perm="auth.can_revoke"),
            render=DummySignoffRenderer(),
        )

    def signed_signoff(self):
        """Return a signed signoff instance"""
        signoff = self.signoff_type()
        signoff.sign(user=self.user)
        return signoff

    def render_template(self, template, **context):
        """Render test template with context overrides"""
        context = context or {}
        defaults = dict(
            signoff=self.signoff_type(),
            request_user=self.user,
        )
        return template.render(Context({**defaults, **context}))

    def test_render_signoff_form_default(self):
        """Unsigned signoff renders form by default"""
        out = self.render_template(
            Template("{% load signoff_tags %}{% render_signoff signoff %}")
        )
        self.assertEqual(out, f"Signoff Form for {self.signoff_type.id}")

    def test_render_signoff_form_explicit(self):
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{% render_signoff signoff action='form' %}"
            )
        )
        self.assertEqual(out, f"Signoff Form for {self.signoff_type.id}")

    def test_render_signoff_signet_unsigned(self):
        """Test rendering an unsigned Signet"""
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{% render_signoff signoff action='signet' %}"
            )
        )
        self.assertEqual(out, "")

    def test_render_signoff_signet_default(self):
        """Signed signoff renders signet by default"""
        signoff = self.signed_signoff()
        out = self.render_template(
            Template("{% load signoff_tags %}{% render_signoff signoff %}"),
            signoff=signoff,
        )
        self.assertEqual(out, f"Signed Signet for {signoff.id}")

    def test_render_signoff_signet_explicit(self):
        signoff = self.signed_signoff()
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{% render_signoff signoff action='signet' %}"
            ),
            signoff=signoff,
        )
        self.assertEqual(out, f"Signed Signet for {signoff.id}")

    def test_render_signoff_form_signed(self):
        """Test rendering form for a signed Signet"""
        signoff = self.signed_signoff()
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{% render_signoff signoff action='form' %}"
            ),
            signoff=signoff,
        )
        self.assertEqual(out, "")

    def test_render_signoff_signet_custom_template(self):
        """For fine control over signet rendering - pass template name in tag"""
        signoff = signoffs.agree_signoff()
        signoff.sign(user=self.user)
        out = self.render_template(
            Template(
                "{% load signoff_tags %}"
                "{% render_signoff signoff signet_template='testapp/signets/custom_signet.html' %}"
            ),
            signoff=signoff,
        )
        self.assertEqual(out.strip(), f"Signoff signed by {self.user.username}")

    def test_render_signoff_form_custom_template(self):
        """For fine control over signoff form rendering - pass template name in tag"""
        signoff = signoffs.agree_signoff()
        out = self.render_template(
            Template(
                "{% load signoff_tags %}"
                "{% render_signoff signoff signoff_form_template='testapp/signets/custom_signet.html' %}"
            ),
            signoff=signoff,
        )
        self.assertEqual(out.strip(), f"Signoff { signoff.id } Form")

    def test_can_revoke_false(self):
        signoff = self.signed_signoff()
        out = self.render_template(
            Template("{% load signoff_tags %}{{ signoff|can_revoke:request_user }}"),
            signoff=signoff,
        )
        self.assertEqual(out, "False")

    def test_can_revoke_true(self):
        revoke = fixtures.get_perm("can_revoke")
        user = fixtures.get_user(perms=(revoke,))
        signoff = self.signed_signoff()
        out = self.render_template(
            Template("{% load signoff_tags %}{{ signoff|can_revoke:request_user }}"),
            signoff=signoff,
            request_user=user,
        )
        self.assertEqual(out, "True")


class DummyApprovalInstanceRenderer(ApprovalInstanceRenderer):
    """Dummy renderer with easy-to-test presentation logic"""

    def __call__(self, request_user=None, context=None, **kwargs):
        """Return a string containing a rendered version of this approval, optionally tailored for requesting user."""
        return (
            f"Approval {self.approval.id} Approved with {self.approval.signoffs.count()} signoffs"
            if self.approval.is_approved()
            else f"Approval {self.approval.id} with {self.approval.signoffs.count()} signoffs"
        )


DummyApprovalRenderer = utils.service(DummyApprovalInstanceRenderer)


class RenderApprovalTagTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = fixtures.get_user()
        cls.signoff_type = signoff_type = ApprovalSignoff.register(
            id="test.approval.signoff"
        )
        _ = fixtures.get_perm("can_revoke")
        cls.approval_type = SimpleApproval.register(
            id="test.approval",
            logic=ApprovalLogic(revoke_perm="auth.can_revoke"),
            render=DummyApprovalRenderer(),
            signing_order=signing_order.SigningOrder(
                signing_order.OneOrMore(signoff_type)
            ),
        )

    def setUp(self):
        super().setUp()
        self.approval = self.approval_type.create()

    def signed_signoff(self):
        """Return a signed signoff instance on the approval"""
        signoff = self.approval.get_next_signoff(for_user=self.user)
        signoff.sign(user=self.user)
        return signoff

    def render_template(self, template, **context):
        """Render test template with context overrides"""
        context = context or {}
        defaults = dict(
            approval=self.approval,
            request_user=self.user,
        )
        return template.render(Context({**defaults, **context}))

    def test_render_approval_empty(self):
        """Unsigned signoff renders form by default"""
        out = self.render_template(
            Template("{% load signoff_tags %}{% render_approval approval %}")
        )
        self.assertEqual(out, f"Approval {self.approval_type.id} with 0 signoffs")

    def test_render_approval_signed(self):
        """Unsigned signoff renders form by default"""
        for i in range(1, 4):
            self.signed_signoff()
            out = self.render_template(
                Template("{% load signoff_tags %}{% render_approval approval %}")
            )
            self.assertEqual(out, f"Approval {self.approval_type.id} with {i} signoffs")

    def test_render_approval_approved(self):
        """Unsigned signoff renders form by default"""
        self.signed_signoff()
        self.approval.approve()
        out = self.render_template(
            Template("{% load signoff_tags %}{% render_approval approval %}")
        )
        self.assertEqual(
            out, f"Approval {self.approval_type.id} Approved with 1 signoffs"
        )

    def test_can_revoke_false(self):
        self.signed_signoff()
        self.approval.approve()
        out = self.render_template(
            Template("{% load signoff_tags %}{{ approval|can_revoke:request_user }}"),
        )
        self.assertEqual(out, "False")

    def test_can_revoke_true(self):
        revoke = fixtures.get_perm("can_revoke")
        user = fixtures.get_user(perms=(revoke,))
        self.signed_signoff()
        self.approval.approve()
        out = self.render_template(
            Template("{% load signoff_tags %}{{ approval|can_revoke:request_user }}"),
            request_user=user,
        )
        self.assertEqual(out, "True")

    def test_next_signoff_for_user(self):
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{{ approval|next_signoff_for_user:request_user }}"
            ),
        )
        self.assertEqual(out, "test.approval.signoff (unsigned)")

    def test_next_signoffs_for_user(self):
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{{ approval|next_signoffs_for_user:request_user|length }}"
            ),
        )
        self.assertEqual(out, "1")

    def test_next_signoff_for_user_none(self):
        self.signed_signoff()
        self.approval.approve()
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{{ approval|next_signoff_for_user:request_user }}"
            ),
        )
        self.assertEqual(out, "None")

    def test_next_signoffs_for_user_none(self):
        self.signed_signoff()
        self.approval.approve()
        out = self.render_template(
            Template(
                "{% load signoff_tags %}{{ approval|next_signoffs_for_user:request_user|length }}"
            ),
        )
        self.assertEqual(out, "0")



"""
    Signoff sequence ordering automation, based on pattern matching Signoff instances to expected Types.
"""
from typing import Protocol

from django.core.exceptions import ImproperlyConfigured

from signoffs.core.signoffs import AbstractSignoff

from . import signoff_pattern as pm


def validate_signing_order_pattern(pattern):
    """Raise Improperly configured if the given signing order pattern is not valid."""
    terms = list(pattern.terms())
    if not len(terms) > 0 and not all(issubclass(AbstractSignoff, t) for t in terms):
        raise ImproperlyConfigured(
            "SigningOrder: at least one pattern term required; all terms must be Signoff Types"
        )
    # The following constraint is not needed in theory, but is a practical and performance reality
    if not all(
        a.get_signetModel() == b.get_signetModel() for a, b in zip(terms, terms[1:])
    ):
        raise ImproperlyConfigured(
            "SigningOrder: all pattern Signoff Types must share the same Signet model."
        )
    return True


class SigningOrderStrategyProtocol(Protocol):
    """Protocol for defining the API required to define a strategy for ordering a sequence of signoffs"""

    def next_signoffs(self) -> list:
        """Return a list of the next Signoff Type(s) available for signing in this signing order"""
        ...

    def is_complete(self) -> bool:
        """Return True iff this signing order is complete (all required signoffs have been signed)"""
        ...


class SigningOrderPatternMatcher:
    """
    Ordering Strategy: match a pattern of Signoff Types defining a "signing order" against a signets queryset.

    Implements SigningOrderStrategyProtocol
    match result object can answer questions like:
        - does the sequence of signoffs satisfy the pattern defined; and
        - what SignoffType(s) could be added to the sequence next?
    """

    def __init__(self, pattern: pm.SigningOrderPattern, signets_queryset):
        """
        Match the signets queryset against the Singing Order pattern
        signet_set must be ordered chronologically (by timestamp), which is default ordering for Signet
        """
        validate_signing_order_pattern(pattern)
        self.pattern = pattern
        self.signets_queryset = signets_queryset

    @property
    def match(self):
        """Return a pm.MatchResult object for matching pattern against queryset (lazy evaluation)"""
        return self.pattern.match(*list(self.signets_queryset))

    def next_signoffs(self) -> list:
        """Return a list of the next Signoff Type(s) available for signing in this signing order"""
        return self.match.next

    def is_complete(self) -> bool:
        """Return True iff this signing order is complete (all required signoffs have been signed)"""
        return self.match.is_complete


class SigningOrder:
    """
    A descriptor used to "inject" a SigningOrder Strategy object into its owner's instances.
    The strategy_class provides a strategy for sequencing the owner's signoffs (signet_set).
    """

    default_strategy_class = SigningOrderPatternMatcher  # default service is a general-purpose pattern matcher

    def __init__(
        self, *pattern, signet_set_accessor="signatories", strategy_class=None
    ):
        """
        This descriptor injects a `SigningOrderStrategyProtocol` object to manage the signing order for the owner's `signet_set`

        `pattern` is a sequence of `SigningOrderPattern` objects (or Signoff Types, which are also legal Pattern tokens)
            defining the signing order, typically for an Approval, but any kind of owner object.
        `pattern` is passed directly through to the `strategy_class` constructor, so could, in theory, be anything.
        `signet_set_accessor` is string with name of callable or attribute for a `Signet` manager on that owner instance.
        `strategy_class` allows this descriptor to be re-used with other ordering strategies
        """
        pattern = pm.InSeries(*pattern)
        validate_signing_order_pattern(pattern)
        self.pattern = pattern
        self.signet_set_accessor = signet_set_accessor
        self.strategy_class = strategy_class or self.default_strategy_class

    def get_service_instance(self, owner_instance):
        """Return an instance of the `strategy_class` for the given owner instance"""
        signet_set_accessor = getattr(owner_instance, self.signet_set_accessor)
        return self.strategy_class(
            pattern=self.pattern, signets_queryset=signet_set_accessor.all()
        )

    def __get__(self, instance, owner=None):
        """
        Instantiate and return a `strategy_class` instance to provide SigningOrder services for the owning instance
        """
        if instance is None:  # class access - nada - nothing useful?
            return self
        else:  # instance access - return the ordering strategy object
            return self.get_service_instance(instance)


__all__ = [
    "SigningOrder",
    "SigningOrderStrategyProtocol",
    "SigningOrderPatternMatcher",
]



""" Default settings can be overridden in project settings """

from django.conf import settings

# Autodiscover all 'signoffs' modules in installed apps
# Module name to autodiscover 'signoffs' -  None to disable (e.g. if signoffs are defined in models.py)
SIGNOFFS_AUTODISCOVER_MODULE = getattr(
    settings, "SIGNOFFS_AUTODISCOVER_MODULE", "signoffs"
)

# dictionary or callable(signet) that returns default values for mutable signet fields.
SIGNOFFS_SIGNET_DEFAULTS = getattr(settings, "SIGNOFFS_SIGNET_DEFAULTS", None)

# SIGNOFFS_SETTING = getattr(settings, 'SIGNOFFS_SETTING', 'DEFAULT')



"""
    Some basic Signoff Types backed by the Signet models defined in this package

    Add `"signoffs.contrib.signets"` to `settings.INSTALLED_APPS`
"""

from signoffs.core.signoffs import BaseSignoff, SignoffLogic
from signoffs.registry import register

from .models import RevokedSignet, Signet


@register(id="signoffs.simple-signoff")
class SimpleSignoff(BaseSignoff):
    """
    A basic Signoff Type that can be used out-of-the-box for simple use-cases where any user can sign off

    Uses `DefaultSignoffBusinessLogic` - unrestricted: anyone can sign or revoke.
    Backed by `signoffs.contrib.signets.models.Signet` model.
    """

    signetModel = Signet
    revokeModel = None  # revoking a SimpleSignoff just deletes it
    label = "I consent"


@register(id="signoffs.revokable-signoff")
class RevokableSignoff(SimpleSignoff):
    """A SimpleSignoff that stores a "receipt" when a signoff is revoked"""

    revokeModel = RevokedSignet


@register(id="signoffs.irrevokable-signoff")
class IrrevokableSignoff(SimpleSignoff):
    """A SimpleSignoff that can never be revoked"""

    revoke_perm = False
    logic = SignoffLogic(perm=None, revoke_perm=False)  # restrict revoke



"""
App-dependent tests for signoff model fields and relation descriptors
"""
from django.test import TestCase

from signoffs.core.tests import fixtures
from tests.test_app import models, signoffs


class SignoffSetTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.employee = fixtures.get_user()
        cls.hr1 = fixtures.get_user(
            first_name="Herman",
            last_name="Roy",
            username="hr1",
            perms=("can_approve_hr",),
        )
        cls.hr2 = fixtures.get_user(
            first_name="Hillary",
            last_name="Rice",
            username="hr2",
            perms=("can_approve_hr",),
        )
        cls.vacation = models.Vacation.objects.create(employee="Bob")
        # Signets can be added via the class...
        s1 = models.Vacation.signoffset.create(user=cls.hr1, vacation=cls.vacation)
        # or directly on the related instance
        s2 = cls.vacation.signoffset.create(user=cls.hr2)
        cls.signoffs = [s1, s2]
        cls.signets = [s.signet for s in cls.signoffs]

    def test_signoffset(self):
        self.assertEqual(models.Vacation.signoffset, signoffs.hr_signoff)
        self.assertListEqual(self.vacation.signoffset.all(), self.signoffs)


# TODO: extend these tests to exercise signoffset and signofffield, verify sigil field, revoke logic, etc.
#       test revoking a SignoffField - does this break DB constraint?



"""
    Some basic Approval Types backed by the Stamp model defined in this package.
"""

from signoffs.core.approvals import BaseApproval
from signoffs.core.forms import SignoffFormsManager
from signoffs.core.signoffs import BaseSignoff
from signoffs.registry import register

from .models import Signet as ApprovalSignet
from .models import Stamp


def approval_signoff_form():
    """Avoid circular imports that might arise from importing form before models are finished loading"""
    from . import forms

    return forms.ApprovalSignoffForm


class ApprovalSignoff(BaseSignoff):
    """An abstract, base Signoff Type backed by a ApprovalSignet - a Signet with a FK relation to an ApprovalStamp"""

    signetModel = ApprovalSignet

    forms = SignoffFormsManager(signoff_form=approval_signoff_form)

    @property
    def subject(self):
        """Subject is the approval being signed off on."""
        return self._subject or self.signet.stamp.approval

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def approval(self):
        """friendly name for subject"""
        return self.subject


@register(id="signoffs.simple-approval")
class SimpleApproval(BaseApproval):
    """
    A base Approval Type that can be used out-of-the-box for simple use-cases where any user can sign off
    Backed by signoffs.contrib.approvals.models.Stamp model.
    """

    stampModel = Stamp
    label = "Approve"


@register(id="signoffs.irrevokable-approval")
class IrrevokableApproval(SimpleApproval):
    """A SimpleApproval that can never be revoked"""

    revoke_perm = False



"""
    Custom object and query managers.
"""
from signoffs import registry


class QuerySetApiMixin:
    """
    Delegates common methods to a query manager or queryset to emulate a queryset-like API,
        without making extra queries where possible.  Assumes a small number of instances on the queryset
    """

    qs = None  # queryset or query manager, to be defined on mixed-in instance

    def all(self):
        """Return list of objects in this set, ordered chronologically"""
        return self.qs.all()

    def count(self):
        """Return the number of objects in the qs"""
        return len(self.all())

    def exists(self):
        """Rerturn True iff at least one Object exists in this set"""
        return bool(self.all())

    def earliest(self):
        """Return the first object from this set, or None if not self.exists()"""
        all = self.all()
        return all[0] if all else None

    def latest(self):
        """Return most recent object from this set, or None if not self.exists()"""
        all = self.all()
        return all[-1] if all else None


# Signoff / Signet Sets


class SignetSetApiMixin(QuerySetApiMixin):
    """Delegates common methods to a signet_set manager or queryset"""

    signet_set = None  # Signet qs or manager, to be defined on mixed-in instance

    @property
    def qs(self):
        return self.signet_set


class SignoffSetManager(SignetSetApiMixin):
    """
    Manage a set of Signoff objects backed by a signet manager or queryset.
    May be used, for example, on the reverse side of a many-to-one relation formed by a Signet with a FK.

    For example::

        signets = Signets.objects.filter(user=bob)
        bobs_report_signoffs = SignoffSetManager('myapp.signoffs.report', signets)

    For "reverse" access from a signet-related object, use a SignoffSet field to define the reverse manager.

    The API is modelled on django's related object managers, with familiar operations that work roughly equivalently.
    """

    def __init__(self, signoff_type, signet_set, signet_set_owner=None):
        """
        Manage the signet_set (query manager or queryset), filtered for the given Signoff Type
        If an approval instance is supplied, signets in this set are created with a reference to the approval's stamp.
        """
        self.signoff_type = registry.get_signoff_type(signoff_type)
        self.signet_set = signet_set
        self.signet_set_owner = signet_set_owner

    # Customize queryset emulation provided by Signets Set API Mixin
    def all(self):
        """Return list of signoffs in this set, ordered chronologically"""
        return super().all().signoffs(signoff_id=self.signoff_type.id)

    def _pre_save_owner(self):
        """Bit of a hack - if signet_set_owner is an unsaved model, save it before saving related signets"""
        try:
            if self.signet_set_owner._meta.model and not self.signet_set_owner.pk:
                self.signet_set_owner.save()
        except AttributeError:
            pass

    def create(self, user, **kwargs):
        """Create and return a new Signoff in this set"""
        self._pre_save_owner()
        signet = self.signet_set.create(
            signoff_id=self.signoff_type.id, user=user, **kwargs
        )
        return self.signoff_type(signet)

    # signoff delegate / aggregate methods

    def can_sign(self, user):
        """Return True iff given user would be allowed to sign (create) a new signoff in this set"""
        return self.signoff_type().can_sign(user)

    def has_signed(self, user):
        """Return True iff given user is a signatory in this set of signoffs"""
        return any(s.signatory == user for s in self.all())

    @property
    def forms(self):
        """
        Return the forms manager for adding a new signoff to this set or revoking a signoff in this set.
        """
        return self.signoff_type.forms

    # related querysets

    def revoked(self):
        """
        Returns a queryset of revoked signets related to the manager's instance.
        related 'revoked' and 'revoked.user' objects are selected, assuming why else get the revoked signoffs
        """
        return (
            self.signoff_type.get_revoked_signets_queryset()
            .filter(**self.signet_set.core_filters)
            .select_related("revoked__user")
        )


class SignoffSingleManager(SignoffSetManager):
    """
    A SignoffSetManager that attempts to limit the number of related signets to one
        and provide simplified API for dealing with the single signoff.
    Why?
        b/c it is convenient to manage all signoffs, e.g., on an Approval, backed by one Signet model.
        If every Signoff is a single, then a set of SignoffFields is well-suited to the task.
        But if any Signoff may be a series, then it is convenient to model all as SignoffSets.
        This class helps clarify that a specific reverse Many2One relation is being modeled as a OneToOne.
    Caveats
        Can't really prevent multiple signoffs being added via non-API access.  That's up to the application programmer.
    """

    def get(self):
        """Get this signoff"""
        return self.all()[0] if self.exists() else self.signoff_type()

    def create(self, user, **kwargs):
        """Create and return the new Signoff. Raise ??? if self.exists()"""
        if self.exists():
            raise Exception  # TODO: what exception here???
        return super().create(user, **kwargs)

    def can_sign(self, user):
        """Return True iff given user would be allowed to sign (create) a new signoff in this set"""
        return not self.exists() and super().can_sign(user)


class StampSignoffsManager(SignetSetApiMixin):
    """
    Manage the entire set of signoffs related to a Stamp of Approval via reverse signet_set manager.
    Note that a single Stamp manages one set of Signets that may be of various Signoff Types
        -  recommend using a SignoffSetManager instead, unless access to entire set of signoffs is needed.
    Provides a unified API for accessing a single Stamp instance's related signoffs and signets.

    For example::

        stamp = ProjectStamp.objects.filter(project_id='bob').first()
        bobs_project_signoffs = StampSignoffsManager(stamp)

    The API is modelled on django's related object managers, with familiar operations that work roughly equivalently.
    """

    def __init__(self, stamp, subject=None):
        """Manage the given Approval instance and all of its related data"""
        self.stamp = stamp
        self.subject = subject

    @property
    def signet_set(self):
        """required for SignetSetApiMixin"""
        return self.stamp.signatories

    # Customize queryset emulation provided by Signets Set API Mixin
    def all(self):
        """Return list of signoffs in this approval, ordered chronologically"""
        return super().all().signoffs(subject=self.subject)


# Approval / Stamp Sets


class ApprovalStampSetApiMixin(QuerySetApiMixin):
    """Delegates common methods to a stamp_set manager or queryset"""

    stamp_set = (
        None  # Stamp queryset or query manager, to be defined on mixed-in instance
    )

    @property
    def qs(self):
        return self.stamp_set


class ApprovalSetManager(ApprovalStampSetApiMixin):
    """
    Manage a set of Approval objects backed by a Stamp manager or queryset.
    May be used, for example, on the reverse side of a many-to-one relation formed by a Stamp with a FK.
    Provides a unified API for accessing Approval instances, and related SigningOrder, Stamp, signoffs, and signets.

    For example::

        stamps = Stamp.objects.filter(user=bob)
        bobs_report_approvals = ApprovalSetManager('myapp.approvals.report', stamps)

    For "reverse" access from a stamp-related object, use an ApprovalSet field to define the reverse manager.

    The API is modelled on django's related object managers, with familiar operations that work roughly equivalently.
    """

    def __init__(self, approval_type, stamp_set, subject=None):
        """Manage the stamp_set (query manager or queryset), filtered for the given Signoff Type"""
        self.approval_type = registry.get_approval_type(approval_type)
        self.stamp_set = stamp_set
        self.subject = subject

    # Customize queryset emulation provided by Stamp Set API Mixin
    def all(self):
        """Return list of approvals in this set, ordered chronologically"""
        return (
            super()
            .all()
            .approvals(approval_id=self.approval_type.id, subject=self.subject)
        )

    def create(self, **kwargs):
        """Create and return a new Approval in this set"""
        stamp = self.stamp_set.create(approval_id=self.approval_type.id, **kwargs)
        return self.approval_type(stamp, subject=self.subject)



"""
    Basic concrete implementations for ApprovalStamp related models
"""

from signoffs.core.models import (
    AbstractApprovalSignet,
    AbstractApprovalStamp,
    AbstractRevokedSignet,
)


class Signet(AbstractApprovalSignet):
    """A concrete persistence layer for a basic Signet with a relation to an ApprovalStamp"""

    pass


class RevokedSignet(AbstractRevokedSignet):
    """
    A concrete persistence layer for tracking revoked Approval Signets.
    May be declared on Approval Types to provide persistence / tracking of revoked signoffs.
    """

    pass


class Stamp(AbstractApprovalStamp):
    """A concrete persistence layer for basic Approval Types"""

    pass



"""
    Proxy for Approval Types to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""
from django.apps import apps

from signoffs.core import signing_order

from signoffs.core.approvals import (
    AbstractApproval,
    ApprovalLogic,
    BaseApproval,
)
from signoffs.core.renderers import (
    ApprovalInstanceRenderer,
    ApprovalRenderer,
)
from signoffs.core.status import (
    ApprovalInstanceStatus,
    ApprovalStatus,
)
from signoffs.core.urls import (
    ApprovalInstanceUrls,
    ApprovalUrlsManager,
)

if apps.is_installed("signoffs.contrib.approvals"):
    from signoffs.contrib.approvals.approvals import (
        ApprovalSignoff,
        IrrevokableApproval,
        SimpleApproval,
    )



"""
    Custom model fields and relation descriptors
"""
from functools import cached_property
from typing import Type, Union

from django.core.exceptions import ImproperlyConfigured
from django.db import models

from signoffs import registry
from signoffs.core.approvals import AbstractApproval
from signoffs.core.models import AbstractApprovalStamp, AbstractSignet
from signoffs.core.models.managers import (
    ApprovalSetManager,
    SignoffSetManager,
    SignoffSingleManager,
)
from signoffs.core.signoffs import AbstractSignoff
from signoffs.core.utils import Accessor

# OneToOne "forward" relation from arbitrary model to a single Signoff


class RelatedSignoffDescriptor:
    """
    Descriptor that provides a Signoff backed by a OneToOneField to a Signet model,
        with services to manage the relation on the defining model.
    Class access yields the Signoff Type class.
    Instance access yields a Signoff instance wrapping the related Signet object
        (which may be None, i.e. unsigned Signoff)
    """

    def __init__(self, signoff_type, signet_field):
        """Manage a OneToOne signet_field using the given Signoff Type (or signoff id str)"""
        self.signoff_type = signoff_type
        self.signet_field = signet_field
        self.accessor_attr = None  # name of accessor attr; see __set_name__

    def __set_name__(self, owner, name):
        """Grab the field named used by owning class to refer to this descriptor"""
        self.accessor_attr = name

    def _validate_related_model(self, signet_field, signoff_type):
        """Raises ImproperlyConfigured if the signet_field model relation is not same as the signoff_type Signet model"""
        signoff_type = registry.get_signoff_type(signoff_type)
        signet_model = signoff_type.get_signetModel()
        related = signet_field.remote_field.model
        # Related could be a model class or a str - handle either case...
        signet = (
            signet_model._meta.label_lower if isinstance(related, str) else signet_model
        )
        if related != signet:
            raise ImproperlyConfigured(
                f'signet_field "to" model {related} must match the signoff_type Signet model {signet} '
                '- use convenience method "SignoffField" to handle this.'
            )

    def __get__(self, instance, owner=None):
        self._validate_related_model(self.signet_field, self.signoff_type)

        base_signoff_type = registry.get_signoff_type(self.signoff_type)

        class RelatedSignoff(base_signoff_type):
            signet_field = self.signet_field
            """ A Signoff that is aware of a "reverse" one-to-one relation to the instance object """

            def save(self, *args, **kwargs):
                """Save the related signet and then the instance relation"""
                signoff = super().save(*args, **kwargs)
                setattr(instance, self.signet_field.name, signoff.signet)
                instance.save()
                return signoff

        if not instance:
            return RelatedSignoff
        else:
            signoff = RelatedSignoff(
                signet=getattr(instance, self.signet_field.name), subject=instance
            )
            setattr(instance, self.accessor_attr, signoff)
            return signoff


def SignoffField(
    signoff_type, on_delete=models.SET_NULL, null=True, related_name="+", **kwargs
):
    """
    Convenience method for constructing from minimal inputs:
        (1) a sensible OneToOneField(signoff_type.signetModel); and
        (2) an RelatedSignoffDescriptor(signoff_type)
    signoff_type may be an Approval Type or a registered(!) signoff id.
    Default parameter rationale:
        null=True, on_delete=SET_NULL make sensible defaults for a Signet relation since presence/absence is semantic;
            think twice before using other values!
        reverse related_name from Signet generally not so useful, consider disabling with '+'

    In the example::

        from signoffs.signoffs import SimpleSignoff

        applicant_signoff = SimpleSignoff.register('myapp.signoff.applicant', ...)

        # the verbose / explicit way...
        class Application(models.Model):
            # ...
            applicant_signet = OneToOneField(applicant_signoff.signetModel, on_delete=models.SET_NULL,
                                             signoff_type=applicant_signoff,
                                             null=True, )
            applicant_signoff = RelatedSignoffDescriptor(applicant_signoff, applicant_signet)

        # the concise convenient way...
        class Application(models.Model):
            # ...
            applicant_signoff, applicant_signet = SignoffField(applicant_signoff)

    In either case above:
        ``Application.applicant_signet`` is a "normal" One-to-One relation to the Signet model
        ``Application().applicant_signet`` is a Signet instance or None (if signoff is not  signed)

        ``Application.applicant_signoff`` is a RelatedSignoffDescriptor that ultimately injects a Signoff instance
        ``Application().applicant_signoff`` is a Signoff instance of the given signoff_type backed by the applicant_signet
                relation.  This Signoff will update the instance's applicant_signet relation whenever it is save()'ed
    """
    try:
        signoff_type = registry.get_signoff_type(signoff_type)
    except ImproperlyConfigured:
        raise ImproperlyConfigured(
            f"SignoffField: signoff_type {signoff_type} must be registered before it can be used to form a relation. "
            "OneToOneField + RelatedSignoff can form relation to Signet Model with a deferred signoff id."
        )
    # Intentionally using raw .signetModel, (possibly a str: 'app_label.model_name'), so use can avoid circular imports
    signet_field = models.OneToOneField(
        signoff_type.signetModel,
        on_delete=on_delete,
        null=null,
        related_name=related_name,
        **kwargs,
    )
    signoff = RelatedSignoffDescriptor(signoff_type, signet_field)
    return signoff, signet_field


# ManyToOne "reverse" relation from a model instance to the set of Signoff that refer to it.


class SignoffSet:
    """
    A descriptor that injects a "reverse relation" manager, filtered for the specific Signoff type.
    The signoff_type MUST be backed by a Signet with a FK relation to the owning model
    signet_set_accessor is a string with a path to the reverse relation manager on the owning model
        - it may use '__' to travese relations and include method calls e.g.,  'my_signet_manager__get_signet_set'

    In the example::

        class VacationSignet(AbstractSignet):
            vacation = models.ForeignKey(Vacation, on_delete=models.CASCADE, related_name='signatories')

        class Vacation(models.Model):
            employee = models.CharField(max_length=128)
            # ...
            hr_signoffs = SignoffSet('test_app.hr_signoff')
            mngr_signoffs = SignoffSet('test_app.mngr_signoff')


    ``Vacation.signatories`` is the "normal" ``ReverseManyToOneDescriptor`` instance to access the related Signets.
    ``Vacation.hr_signoffs`` and ``Vacation.mngr_signoffs`` are 2 distinct Signoff Types (Signoff sub-classes), while
    ``Vacation().hr_signoffs`` and ``Vacation().mngr_signoffs``  are SignoffSetManager instances
            used to access the set of signoffs of that particular type, backed by the vacation instance.signatories.

    See managers.SignoffSetManager for access API.
    """

    signoff_set_manager = SignoffSetManager

    def __init__(
        self,
        signoff_type: Union[str, Type[AbstractSignoff]] = None,
        signet_set_accessor="signatories",
        signoff_set_manager=None,
        **kwargs,
    ):
        """
        Inject a Signoff; kwargs passed directly through to AbstractSignoff.__init__
        id defaults to the descriptor name if not specified
        """
        self._signoff_type = signoff_type
        self.signet_set_accessor = Accessor(signet_set_accessor)
        self.signoff_set_manager = signoff_set_manager or self.signoff_set_manager
        self.accessor_field_name = ""  # set by __set_name__
        self.kwargs = kwargs

    def __set_name__(self, owner, name):
        """Grab the field named used by owning class to refer to this descriptor"""
        self.accessor_field_name = name

    @staticmethod
    def _is_valid_signoff_type(signoff_type):
        return (
            signoff_type is not None
            and issubclass(signoff_type, AbstractSignoff)
            and issubclass(signoff_type.get_signetModel(), AbstractSignet)
            and signoff_type.has_object_relation()
        )

    @cached_property
    def signoff_type(self):
        """Lazy evaluation for signoff_type to allow all signoffs to register before resolving."""
        signoff_type = registry.get_signoff_type(self._signoff_type)
        if self._is_valid_signoff_type(signoff_type):
            return signoff_type
        else:
            raise ImproperlyConfigured(
                f"SignoffSet: Signoff Type {signoff_type} must have a Signet with a relation"
            )

    def signet_set_owner(self, instance):
        signet_set_owner, _ = self.signet_set_accessor.penultimate(instance)
        return signet_set_owner or instance

    def _validate_related_manager(self, instance):
        """Raises ImproperlyConfigured if the signet_set_accessor is not a relation to a Signet Manager or queryset"""
        try:
            signet_set = self.signet_set_accessor.resolve(instance)
            signet_model = self.signoff_type.get_signetModel()
        except AttributeError:
            raise ImproperlyConfigured(
                f'SignoffSet.signet_set_accessor "{self.signet_set_accessor}" '
                f"does not exist on related model {type(instance)}."
            )

        signet_set_owner = self.signet_set_owner(instance)
        related_models = [
            ro.related_model for ro in signet_set_owner._meta.related_objects
        ]
        if (
            signet_set.model not in related_models
            or signet_set.model is not signet_model
        ):
            raise ImproperlyConfigured(
                f'SignoffSet.signet_set_accessor "{self.signet_set_accessor}" '
                f"must be a related {signet_model} to {type(instance)}."
            )

    def get_signoffs_manager(self, instance):
        """Return a signoff_set_manager for signet set instance.signet_set_accessor"""
        self._validate_related_manager(instance)
        signet_set = self.signet_set_accessor.resolve(instance)
        signet_set_owner = self.signet_set_owner(instance)
        return self.signoff_set_manager(self.signoff_type, signet_set, signet_set_owner)

    def __get__(self, instance, owner=None):
        """
        Use the enclosing instance to construct and return a SignoffSetManager instance,
          and replace descriptor with that object
        """
        if instance is None:  # class access - provide access to the Signoff Type itself
            return self.signoff_type
        else:  # on instance, replace descriptor with related signoff manager.  Voilà!
            signoffs_manager = self.get_signoffs_manager(instance)
            setattr(instance, self.accessor_field_name, signoffs_manager)
            return signoffs_manager


def SignoffSingle(*args, **kwargs):
    """A thin wrapper for SignoffSet that configures the correct SignoffSetManager for a single signoff"""
    default_kwargs = dict(
        signoff_set_manager=SignoffSingleManager,
    )
    default_kwargs.update(kwargs)
    return SignoffSet(*args, **default_kwargs)


def ApprovalSignoffSet(*args, **kwargs):
    """A thin wrapper for SignoffSet that configures the correct signet_set_accessor for Approvals"""
    default_kwargs = dict(
        signet_set_accessor="stamp__signatories",
    )
    default_kwargs.update(kwargs)
    return SignoffSet(*args, **default_kwargs)


def ApprovalSignoffSingle(*args, **kwargs):
    """A thin wrapper for SignoffSet that configures the correct signet_set_accessor for a single Approval signoff"""
    default_kwargs = dict(
        signoff_set_manager=SignoffSingleManager,
        signet_set_accessor="stamp__signatories",
    )
    default_kwargs.update(kwargs)
    return SignoffSet(*args, **default_kwargs)


# OneToOne "forward" relation from arbitrary model to a single Approval


class RelatedApprovalDescriptor:
    """
    Descriptor that provides an Approval backed by a OneToOneField to a Stamp model,
        with services to manage the relation on the defining model.
    Class access yields the Approval Type class.
    Instance access yields an Approval instance wrapping the related Stamp object
        (which will be created if it doesn't yet exist, i.e. uninitiated Approval)
    """

    def __init__(self, approval_type, stamp_field):
        """
        Manage a OneToOne Stamp field  using the given Approval Type or approval id
        """
        self.approval_type = approval_type
        self.stamp_field = stamp_field
        self.accessor_attr = None  # see __set_name__

    def __set_name__(self, owner, name):
        """Grab the field named used by owning class to refer to this descriptor"""
        self.accessor_attr = name

    def _validate_related_model(self, stamp_field, approval_type):
        """Raises ImproperlyConfigured if the stamp_field model relation is not same as the approval_type Stamp model"""
        approval_type = registry.get_approval_type(approval_type)
        stamp_model = approval_type.get_stampModel()
        related = stamp_field.remote_field.model
        # Related could be a model class or a str - handle either case...
        stamp = (
            stamp_model._meta.label_lower if isinstance(related, str) else stamp_model
        )
        if related != stamp:
            raise ImproperlyConfigured(
                f'stamp_field "to" model {related} must match the approval_type Stamp model {stamp} '
                '- use convenience method "ApprovalField" to handle this.'
            )

    def __get__(self, instance, owner=None):
        """Return an approval obj. wrapping the stamp_field on instance access, or the approval_type"""
        # Can't validate until models are fully loaded :-/
        self._validate_related_model(self.stamp_field, self.approval_type)

        base_approval_type = registry.get_approval_type(self.approval_type)

        class BaseRelatedApproval(base_approval_type):
            """An Approval that is aware of a "reverse" one-to-one relation to the instance object"""

            stamp_field = self.stamp_field

            def save(self, *args, **kwargs):
                """Save the related stamp and then the instance relation"""
                approval = super().save(*args, **kwargs)
                if not getattr(instance, self.stamp_field.name) == approval.stamp:
                    setattr(instance, self.stamp_field.name, approval.stamp)
                    instance.save()
                return approval

        RelatedApproval = type(
            f"Related{base_approval_type.__name__}", (BaseRelatedApproval,), {}
        )

        if not instance:
            return base_approval_type
        else:
            stamp = getattr(instance, self.stamp_field.name)
            approval = RelatedApproval(stamp=stamp, subject=instance)
            if not stamp:
                approval.save()
            setattr(instance, self.accessor_attr, approval)
            return approval


def ApprovalField(
    approval_type, on_delete=models.SET_NULL, null=True, related_name="+", **kwargs
):
    """
    Convenience method for constructing from minimal inputs:
        (1) a sensible OneToOneField(approval_type.stampModel); and
        (2) an RelatedApprovalDescriptor(approval_type)
    approval_type may be an Approval Type or a registered(!) approval id.
    Default parameter rationale:
        null=True, on_delete=SET_NULL make sensible defaults for a Stamp relation since presence/absence is semantic;
            think twice before using other values!
        related_name defines "reverse relation" from Stamp to the approval subject (object declaring the ApprovalField)
            this is not used internally, but can be very useful e.g., when approval permissions need context of subject
            Wanrning: the name of this field needs to be unquie for each ApprovalField

    In the example::

        from signoffs.approvals import ApprovalSignoff
        from signoffs.models import ApprovalSignet, Stamp

        @register(id=myapp.application_approval)
        class ApplicationApproval(ApprovalSignoff):
            stampModel = Stamp
            S = ApprovalSignoff

            reg_signoff = S.register('application.approval.signoff.registration', ...)
            deposit_signoff = S.register('application.approval.signoff.deposit', ...)
            final_approval = S.register('application.approval.signoff.final', ...)

            signing_order = SigningOrder(
                pm.InSeries(
                    pm.InParallel(
                        ret_signoff,
                        deposit_signoff,
                    )
                    final_approval,
                )
            )

        # The verbose way:
        class Application(models.Model):
            # ...
            approval_stamp = OneToOneField(ApplicationApproval.stampModel, on_delete=models.SET_NULL, null=True)
            approval = RelatedApprovalDescriptor(ApplicationApproval, approval_stamp)

        # The convenient way:
        class Application(models.Model):
            # ...
            approval, approval_stamp = ApprovalField(ApplicationApproval)

    In both cases above:
        ``Application.approval_stamp`` is a "normal" One-to-One relation to the backing Stamp model
        ``Application().approval_stamp`` is a Stamp instance or None (if approval is not initiated)

        ``Application.approval`` is a RelatedApprovalDescriptor that ultimately injects an ApplicationApproval instance
        ``Application().approval`` is an ApplicationApproval instance, backed by the OneToOne approval_stamp relation.
                This approval will update the application's OneToOne relation whenever it is save()'ed
    """
    try:
        approval_type = registry.get_approval_type(approval_type)
    except ImproperlyConfigured:
        raise ImproperlyConfigured(
            f"ApprovalField: approval_type {approval_type} must be registered before it's used to form a relation. "
            "A OneToOneField + RelatedApproval can be used to form this relation with a deferred approval id."
        )
    # Intentionally using raw .stampModel, (possibly 'app_label.model_name' str), so use can avoid circular imports
    stamp_field = models.OneToOneField(
        approval_type.stampModel,
        on_delete=on_delete,
        null=null,
        related_name=related_name,
        **kwargs,
    )
    approval = RelatedApprovalDescriptor(approval_type, stamp_field)
    return approval, stamp_field


# ManyToOne "reverse" relation from a model instance to the set of Approvals that refer to it.
# EXPERIMENTAL: not entirely clear why one would have a set of approvals of one type for same object?
#               would an "approved" object ever need another "approval"?  Maybe like for Building permits?
#               TODO: the Implementation below should work, but requires test cases before use.


class ApprovalSet:
    """
    A descriptor that injects a "reverse relation" manager, filtered for the specific Approval Type.
    The approval_type MUST be backed by a Stamp with a FK relation to the owning model

    In the example::

        class BuildingPermit(AbstractApprovalStamp):
            building = ForeignKey('Building', related_name='permits')

        @register(id='myapp.building_approval')
        class ConstructionApproval(BaseApproval):
            stampModel = BuildingPermit
            S = ApprovalSignoff

            planning_signoff = S.register('building.approval.signoff.planning', ...)
            electrical_signoff = S.register('building.approval.signoff.electrical', ...)
            plumbing_signoff = S.register('building.approval.signoff.plumbing', ...)
            inspection_signoff = S.register('building.approval.signoff.inspection', ...)
            final_approval = S.register('building.approval.signoff.final', ...)

            signing_order = SigningOrder(
                InSeries(
                    planning_signoff,
                    InParallel(
                        electrical_signoff,
                        plumbing_signoff,
                    )
                    AtLeastN(inspection_signoff, 1),
                    final_approval,
                )
            )

        class Building(Model):
            approvals = ApprovalSet(ConstructionApproval, stamp_set_accessor='permits')


    ``BuildingPermit.signatories`` is a "normal" ``ReverseManyToOneDescriptor`` instance to access the related Signets.
    ``ConstructionApproval().signing_order`` defines a SigningOrderManager to match pattern with the stamp's signets.
    ``Building().approvals`` is an ApprovalSetManager instance providing a unified API for accessing the set
            of approvals, backed by building instance.permits.

    See ApprovalSetManager for access API.
    """

    def __init__(
        self,
        approval_type: Union[str, Type[AbstractApproval]] = None,
        stamp_set_accessor="stamp_set",
        **kwargs,
    ):
        """
        Inject ApprovalSetManager; kwargs passed directly through to ApprovalSetManager.__init__
        approval_type may be an Approval Type class or a registered approval id
        """
        self._approval_type = approval_type
        self.stamp_set_accessor = stamp_set_accessor
        self.accessor_field_name = ""  # set by __set_name__
        self.kwargs = kwargs

    def __set_name__(self, owner, name):
        """Grab the field named used by owning class to refer to this descriptor"""
        self.accessor_field_name = name

    @staticmethod
    def _is_valid_approval_type(approval_type):
        # noinspection PyTypeChecker
        return (
            approval_type is not None
            and issubclass(approval_type, AbstractApproval)
            and issubclass(approval_type.get_stampModel(), AbstractApprovalStamp)
            and approval_type.has_object_relation()
        )

    @cached_property
    def approval_type(self):
        approval_type = registry.get_approval_type(self._approval_type)
        if self._is_valid_approval_type(approval_type):
            return approval_type
        else:
            raise ImproperlyConfigured(
                f"ApprovalField - Approval Type {approval_type} must have a Stamp with a relation"
            )

    def _validate_related_manager(self, instance):
        """Raises ImproperlyConfigured if the stamp_set_accessor is not a relation to a Stamp Manager or queryset"""
        try:
            stamp_set = getattr(instance, self.stamp_set_accessor)
            stamp_model = self.approval_type.get_stampModel()
        except AttributeError:
            raise ImproperlyConfigured(
                f'ApprovalSet.stamp_set_accessor "{self.stamp_set_accessor}" '
                f"does not exist on related model {type(instance)}."
            )
        related_models = [ro.related_model for ro in instance._meta.related_objects]
        if stamp_set.model not in related_models or stamp_set.model is not stamp_model:
            raise ImproperlyConfigured(
                f'ApprovalSet.stamp_set_accessor "{self.stamp_set_accessor}" '
                f"must be a related {stamp_model} to {type(instance)}."
            )

    def __get__(self, instance, owner=None):
        """
        Use the enclosing instance to construct a ApprovalSetManager instance,
          then replace descriptor with that object
        """
        if instance is None:  # class access
            return self.approval_type
        else:  # on instance, replace descriptor with related signoff manager.  Voilà!
            approval_set = getattr(instance, self.stamp_set_accessor)
            approval_manager = ApprovalSetManager(
                self.approval_type, approval_set, subject=instance
            )
            setattr(instance, self.accessor_field_name, approval_manager)
            return approval_manager
